/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:   i_pmac.h
 *
 * Purpose: Header for MX driver for Delta Tau PMAC controllers.  Note that
 *          this is the driver for talking to the PMAC itself.  PMAC-controlled
 *          motors are interfaced via the driver in libmx/d_pmac.c.
 */

#ifndef __I_PMAC_H__
#define __I_PMAC_H__

#include "mx_rec.h"
#include "mx_motor.h"
#include "mx_gener.h"
#include "mx_rs232.h"

/* Define the data structures used by a PMAC controller. */

#define MX_NUM_PMAC_MOTORS	8

#define MX_PMAC_RS232_SUBTYPE		1
#define MX_PMAC_EPICS_ECT_SUBTYPE	2

#define MX_PMAC_SUBTYPE_NAME_LENGTH	32
#define MX_PMAC_INTERFACE_ARGS_LENGTH	80

typedef struct {
	MX_RECORD *record;
	int interface_subtype;

	/* PMAC controller PROM version. */

	int major_version;
	int minor_version;
	char subminor_version;

	int i3_variable;  /* affects command-response handshaking */
	int i6_variable;  /* affects error reporting */

	/* Parameters shared by all eight motor axes. */

	MX_RECORD *interface_record;

	char subtype_name[MX_PMAC_SUBTYPE_NAME_LENGTH+1];
	char interface_args[MX_PMAC_INTERFACE_ARGS_LENGTH+1];

	/* The following is an array of pointers to all the individual
	 * motors controlled by this PMAC controller.
	 */

	MX_MOTOR *motor[MX_NUM_PMAC_MOTORS];
} MX_PMAC;

#define MXI_PMAC_STANDARD_FIELDS \
  {-1, -1, "subtype_name", MXFT_STRING, NULL, 1,{MX_PMAC_SUBTYPE_NAME_LENGTH},\
	MXF_REC_TYPE_STRUCT, offsetof(MX_PMAC, subtype_name), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "interface_args", MXFT_STRING, NULL, \
				1, {MX_PMAC_INTERFACE_ARGS_LENGTH},\
	MXF_REC_TYPE_STRUCT, offsetof(MX_PMAC, interface_args), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}

#ifndef MX_CR
#define MX_ACK		'\006'
#define MX_BELL		'\007'
#define MX_LF		'\012'
#define MX_CR		'\015'
#define MX_CTRL_Z	'\032'
#endif

MX_API mx_status_type mxi_pmac_initialize_type( long type );
MX_API mx_status_type mxi_pmac_create_record_structures( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_finish_record_initialization(
							MX_RECORD *record );
MX_API mx_status_type mxi_pmac_delete_record( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_read_parms_from_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_write_parms_to_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_open( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_close( MX_RECORD *record );

MX_API mx_status_type mxi_pmac_getchar( MX_GENERIC *generic,
					char *c, int flags );
MX_API mx_status_type mxi_pmac_putchar( MX_GENERIC *generic,
					char c, int flags );
MX_API mx_status_type mxi_pmac_read( MX_GENERIC *generic,
					void *buffer, size_t count );
MX_API mx_status_type mxi_pmac_write( MX_GENERIC *generic,
					void *buffer, size_t count );
MX_API mx_status_type mxi_pmac_input_is_available( MX_GENERIC *generic,
						int *input_available );
MX_API mx_status_type mxi_pmac_discard_unread_input( MX_GENERIC *generic,
							int debug_flag );
MX_API mx_status_type mxi_pmac_discard_unwritten_output( MX_GENERIC *generic,
							int debug_flag );

extern MX_RECORD_FUNCTION_LIST mxi_pmac_record_function_list;
extern MX_GENERIC_FUNCTION_LIST mxi_pmac_generic_function_list;

extern long mxi_pmac_num_record_fields;
extern MX_RECORD_FIELD_DEFAULTS *mxi_pmac_rfield_def_ptr;

/* === Driver specific functions === */

MX_API mx_status_type mxi_pmac_command( MX_PMAC *pmac, char *command,
	char *response, int response_buffer_length,
	char *error_message, int error_message_buffer_length,
	int debug_flag );
MX_API mx_status_type mxi_pmac_getline( MX_PMAC *pmac,
			char *buffer, long buffer_size, int debug_flag );
MX_API mx_status_type mxi_pmac_putline( MX_PMAC *pmac,
			char *buffer, int debug_flag );
MX_API mx_status_type mxi_pmac_getc( MX_PMAC *pmac,
			char *c, int debug_flag );
MX_API mx_status_type mxi_pmac_putc( MX_PMAC *pmac,
			char c, int debug_flag );

#endif /* __I_PMAC_H__ */
