/*
 * Name:    mx_clock.c
 *
 * Purpose: MX time keeping functions.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#ifndef OS_WIN32
#include <sys/times.h>
#endif

#ifdef OS_WIN32
#include <windows.h>
#endif

#include "mx_util.h"
#include "mx_clock.h"

/* The most recent value of the system clock tick value for this process
 * is kept in the variable mx_most_recent_clock_tick_value.
 */

static MX_CLOCK_TICK mx_most_recent_clock_tick_value;

/* mx_clock_tick_divisor contains 1 + the maximum possible range
 * of values that the low order part of the MX_CLOCK_TICK structure
 * can have converted to a double.  You can use this value to convert
 * a clock tick value represented as a double into the individual
 * high order and low order words of the MX_CLOCK_TICK structure.
 */

static double mx_clock_tick_divisor;

static clock_t mx_low_order_max;
static clock_t mx_low_order_min;

/*-----------------------------------------------------------*/

static clock_t
mx_current_cpu_time( void )
{
	clock_t current_cpu_time;

#if defined(OS_LINUX) || defined(OS_SOLARIS2) || defined(OS_SGI) \
    || defined(OS_HPUX) || defined(OS_DJGPP)

	struct tms buf;

	current_cpu_time = times(&buf);

#elif defined(OS_SUNOS4) || defined(OS_VMS)

	current_cpu_time = clock();

#elif defined(OS_WIN32)

	current_cpu_time = (clock_t) timeGetTime();

#else
#error mx_current_cpu_time() has not been defined for this operating system.
#endif
	return current_cpu_time;
}

/*-----------------------------------------------------------*/

MX_EXPORT double
mx_clock_ticks_per_second( void )
{
	double clock_ticks_per_second;

#if defined(OS_LINUX) || defined(OS_SOLARIS2) || defined(OS_SGI) \
    || defined(OS_HPUX) || defined(OS_VMS)

	clock_ticks_per_second = CLK_TCK;

#elif defined(OS_DJGPP)

	clock_ticks_per_second = CLOCKS_PER_SEC;

#elif defined(OS_WIN32)

	clock_ticks_per_second = 1000.0;

#elif defined(OS_SUNOS4)

	clock_ticks_per_second = 1000000.0;

#else
#error mx_clock_ticks_per_second() has not been defined for this operating system.
#endif

	return clock_ticks_per_second;
}

MX_EXPORT void
mx_initialize_clock_ticks( void )
{
	mx_most_recent_clock_tick_value.high_order = 0L;
	mx_most_recent_clock_tick_value.low_order  = mx_current_cpu_time();

	if ( sizeof(clock_t) == sizeof(long) ) {
		mx_clock_tick_divisor = 1.0 + (double) ULONG_MAX;

		if ( ULONG_MAX == LONG_MAX ) {
			mx_low_order_max = ULONG_MAX;
			mx_low_order_min = 0;
		} else {
			mx_low_order_max = LONG_MAX;
			mx_low_order_min = LONG_MIN;
		}

	} else if ( sizeof(clock_t) == sizeof(int) ) {
		mx_clock_tick_divisor = 1.0 + (double) UINT_MAX;

		if ( UINT_MAX == INT_MAX ) {
			mx_low_order_max = UINT_MAX;
			mx_low_order_min = 0;
		} else {
			mx_low_order_max = INT_MAX;
			mx_low_order_min = INT_MIN;
		}

	} else {
		fprintf(stderr,
	"ERROR: mx_clock_tick_divisor has not been correctly initialized\n"
	"in mx_initialize_clock_ticks().  Aborting...\n" );

		exit(1);
	}
	return;
}

/* mx_current_clock_tick() must be invoked from time to time in order to
 * not miss two rollovers of the low order value in a row.  Currently,
 * the platforms with the most stringent requirements on this are Solaris 2
 * and SGI where if you go more than 2147 seconds (about 36 minutes) 
 * between calls to mx_current_clock_tick(), the value returned by 
 * mx_current_cpu_time() may roll over twice without this being detected
 * by the software.
 */

MX_EXPORT MX_CLOCK_TICK
mx_current_clock_tick( void )
{
	MX_CLOCK_TICK current_clock_tick;
	clock_t current_cpu_time;

	current_clock_tick = mx_most_recent_clock_tick_value;

	current_cpu_time = mx_current_cpu_time();

	/* Check for CPU time rollover. */

	if ( current_cpu_time < current_clock_tick.low_order )
		current_clock_tick.high_order ++;

	current_clock_tick.low_order = current_cpu_time;

	mx_most_recent_clock_tick_value = current_clock_tick;

	return current_clock_tick;
}

MX_EXPORT MX_CLOCK_TICK
mx_convert_seconds_to_clock_ticks( double seconds )
{
	MX_CLOCK_TICK clock_tick_value;
	double clock_ticks_as_double;

	clock_ticks_as_double = seconds * mx_clock_ticks_per_second();

	clock_tick_value.high_order = (unsigned long)
			( clock_ticks_as_double / mx_clock_tick_divisor );

	clock_tick_value.low_order = (clock_t) ( clock_ticks_as_double
	    - mx_clock_tick_divisor * (double) clock_tick_value.high_order );

	return clock_tick_value;
}

MX_EXPORT double
mx_convert_clock_ticks_to_seconds( MX_CLOCK_TICK clock_tick_value )
{
	double clock_ticks_as_double, seconds;

	clock_ticks_as_double = (double) clock_tick_value.low_order
	    + mx_clock_tick_divisor * (double) clock_tick_value.high_order;

	seconds = clock_ticks_as_double / mx_clock_ticks_per_second();

	return seconds;
}

MX_EXPORT MX_CLOCK_TICK
mx_add_clock_ticks( MX_CLOCK_TICK clock_tick_1, MX_CLOCK_TICK clock_tick_2 )
{
	MX_CLOCK_TICK result;
	unsigned long H1, H2;
	clock_t L1, L2;

	H1 = clock_tick_1.high_order;
	L1 = clock_tick_1.low_order;

	H2 = clock_tick_2.high_order;
	L2 = clock_tick_2.low_order;

	result.low_order  = L1 + L2;

	result.high_order = H1 + H2;

	/* Check for carry. */

	if ( L2 > mx_low_order_max - L1 )
		result.high_order ++;

	return result;
}

MX_EXPORT MX_CLOCK_TICK
mx_subtract_clock_ticks(MX_CLOCK_TICK clock_tick_1, MX_CLOCK_TICK clock_tick_2)
{
	MX_CLOCK_TICK result;
	unsigned long H1, H2;
	clock_t L1, L2;

	H1 = clock_tick_1.high_order;
	L1 = clock_tick_1.low_order;

	H2 = clock_tick_2.high_order;
	L2 = clock_tick_2.low_order;

	result.low_order  = L1 - L2;

	result.high_order = H1 - H2;

	/* Check for borrow. */

	if ( L1 < L2 )
		result.high_order --;

	return result;
}

MX_EXPORT int
mx_compare_clock_ticks(MX_CLOCK_TICK clock_tick_1, MX_CLOCK_TICK clock_tick_2)
{
	unsigned long H1, H2;
	clock_t L1, L2;

	H1 = clock_tick_1.high_order;
	L1 = clock_tick_1.low_order;

	H2 = clock_tick_2.high_order;
	L2 = clock_tick_2.low_order;

	if ( H1 < H2 ) {
		return (-1);

	} else if ( H1 > H2 ) {
		return 1;

	} else {
		if ( L1 < L2 ) {
			return (-1);

		} else if ( L1 > L2 ) {
			return 1;

		} else {
			return 0;
		}
	}
}

