/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_epics.c
 *
 * Purpose: Some utilities for using EPICS with MX.
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

/* The following three headers from EPICS. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

#include "mx_rec.h"
#include "mx_epics.h"

MX_EXPORT mx_status_type
mx_epics_convert_mx_type_to_epics_type( long mx_type, chtype *epics_type )
{
	const char fname[] = "mx_epics_convert_mx_type_to_epics_type()";

	switch( mx_type ) {
	case MXFT_STRING:
		*epics_type = DBR_STRING;
		break;
	case MXFT_CHAR:
	case MXFT_UCHAR:
		*epics_type = DBR_CHAR;
		break;
	case MXFT_SHORT:
	case MXFT_USHORT:
		*epics_type = DBR_SHORT;
		break;
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
		*epics_type = DBR_LONG;
		break;
	case MXFT_FLOAT:
		*epics_type = DBR_FLOAT;
		break;
	case MXFT_DOUBLE:
		*epics_type = DBR_DOUBLE;
		break;
	default:
		*epics_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX record field type of %ld cannot be converted to an EPICS type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_epics_type_to_mx_type( chtype epics_type, long *mx_type )
{
	const char fname[] = "mx_epics_convert_epics_type_to_mx_type()";

	switch( epics_type ) {
	case DBR_STRING:
		*mx_type = MXFT_STRING;
		break;
	case DBR_CHAR:
		*mx_type = MXFT_CHAR;
		break;
	case DBR_SHORT:
		*mx_type = MXFT_SHORT;
		break;
	case DBR_LONG:
		*mx_type = MXFT_LONG;
		break;
	case DBR_FLOAT:
		*mx_type = MXFT_FLOAT;
		break;
	case DBR_DOUBLE:
		*mx_type = MXFT_DOUBLE;
		break;
	default:
		*mx_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EPICS type of %ld cannot be converted to an MX record field type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_mx_type_to_ezca_type( long mx_type, char *ezca_type )
{
	const char fname[] = "mx_epics_convert_mx_type_to_ezca_type()";

	switch( mx_type ) {
	case MXFT_STRING:
		*ezca_type = ezcaString;
		break;
	case MXFT_CHAR:
	case MXFT_UCHAR:
		*ezca_type = ezcaByte;
		break;
	case MXFT_SHORT:
	case MXFT_USHORT:
		*ezca_type = ezcaShort;
		break;
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
		*ezca_type = ezcaLong;
		break;
	case MXFT_FLOAT:
		*ezca_type = ezcaFloat;
		break;
	case MXFT_DOUBLE:
		*ezca_type = ezcaDouble;
		break;
	default:
		*ezca_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX record field type of %ld cannot be converted to an EZCA type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_ezca_type_to_mx_type( char ezca_type, long *mx_type )
{
	const char fname[] = "mx_epics_convert_ezca_type_to_mx_type()";

	switch( ezca_type ) {
	case ezcaString:
		*mx_type = MXFT_STRING;
		break;
	case ezcaByte:
		*mx_type = MXFT_CHAR;
		break;
	case ezcaShort:
		*mx_type = MXFT_SHORT;
		break;
	case ezcaLong:
		*mx_type = MXFT_LONG;
		break;
	case ezcaFloat:
		*mx_type = MXFT_FLOAT;
		break;
	case ezcaDouble:
		*mx_type = MXFT_DOUBLE;
		break;
	default:
		*mx_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EZCA type of %d cannot be converted to an MX record field type.",
			ezca_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
