/*
 * Name:    sq_struc.c
 *
 * Purpose: Driver for quick scans that use a Struck MCS via EPICS.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_alloc.h"
#include "mx_scan.h"
#include "mx_scqui.h"
#include "sq_struc.h"
#include "d_pmactc.h"
#include "d_epmtr.h"
#include "d_epmcs.h"

/* EPICS include files. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

MX_RECORD_FUNCTION_LIST mxs_struck_quick_scan_record_function_list = {
	mxs_struck_quick_scan_initialize_type,
	mxs_struck_quick_scan_create_record_structures,
	mxs_struck_quick_scan_finish_record_initialization,
	mxs_struck_quick_scan_delete_record,
	mx_quick_scan_print_scan_structure,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function
};

MX_SCAN_FUNCTION_LIST mxs_struck_quick_scan_scan_function_list = {
	mxs_struck_quick_scan_prepare_for_scan_start,
	mxs_struck_quick_scan_execute_scan_body,
	mxs_struck_quick_scan_cleanup_after_scan_end
};

MX_RECORD_FIELD_DEFAULTS mxs_struck_quick_scan_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCAN_STANDARD_FIELDS,
	MX_QUICK_SCAN_STANDARD_FIELDS
};

long mxs_struck_quick_scan_num_record_fields
			= sizeof( mxs_struck_quick_scan_defaults )
			/ sizeof( mxs_struck_quick_scan_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxs_struck_quick_scan_def_ptr
			= &mxs_struck_quick_scan_defaults[0];

#define EPICS_NAME_LENGTH	41

#define FREE_ARRAYS mxs_struck_quick_scan_free_arrays(scan,struck_quick_scan)

static void
mxs_struck_quick_scan_free_arrays( MX_SCAN *scan,
				MX_STRUCK_QUICK_SCAN *struck_quick_scan )
{
	long i;

	if ( scan == NULL )
		return;

	if ( struck_quick_scan == NULL )
		return;

	if ( struck_quick_scan->motor_position_array != NULL ) {
	    for ( i = 0; i < scan->num_motors; i++ ) {
		if ( struck_quick_scan->motor_position_array[i] != NULL ) {
		    free( struck_quick_scan->motor_position_array[i] );

		    struck_quick_scan->motor_position_array[i] = NULL;
		}
	    }
	}
	if ( struck_quick_scan->data_array != NULL ) {
	    for ( i = 0; i < scan->num_input_devices; i++ ) {
		if ( struck_quick_scan->data_array[i] != NULL ) {
		    free( struck_quick_scan->data_array[i] );

		    struck_quick_scan->data_array[i] = NULL;
		}
	    }
	}
}

static mx_status_type
mxs_struck_quick_scan_get_pointers( MX_SCAN *scan,
				MX_QUICK_SCAN **quick_scan,
				MX_STRUCK_QUICK_SCAN **struck_quick_scan,
				const char *calling_fname )
{
	const char fname[] = "mxs_struck_quick_scan_get_pointers()";

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( quick_scan == (MX_QUICK_SCAN **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_QUICK_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( struck_quick_scan == (MX_STRUCK_QUICK_SCAN **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_STRUCK_QUICK_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*quick_scan = (MX_QUICK_SCAN *) (scan->record->record_class_struct);

	if ( *quick_scan == (MX_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_QUICK_SCAN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	*struck_quick_scan = (MX_STRUCK_QUICK_SCAN *)
					scan->record->record_type_struct;

	if ( *struck_quick_scan == (MX_STRUCK_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_STRUCK_QUICK_SCAN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_initialize_type( long type )
{
	return mx_quick_scan_initialize_type( type );
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_create_record_structures( MX_RECORD *record )
{
	const char fname[] =
		"mxs_struck_quick_scan_create_record_structures()";

	MX_SCAN *scan;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	int i;

	/* Allocate memory for the necessary structures. */

	scan = (MX_SCAN *) malloc( sizeof(MX_SCAN) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAN structure." );
	}

	quick_scan = (MX_QUICK_SCAN *) malloc( sizeof(MX_QUICK_SCAN) );

	if ( quick_scan == (MX_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_QUICK_SCAN structure." );
	}

	struck_quick_scan = (MX_STRUCK_QUICK_SCAN *)
				malloc( sizeof(MX_STRUCK_QUICK_SCAN) );

	if ( struck_quick_scan == (MX_STRUCK_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_STRUCK_QUICK_SCAN structure.");
	}

	/* Now set up the necessary pointers. */

	record->record_superclass_struct = scan;
	record->record_class_struct = quick_scan;
	record->record_type_struct = struck_quick_scan;

	scan->record = record;

	scan->measurement.scan = scan;
	scan->datafile.scan = scan;
	scan->plot.scan = scan;

	quick_scan->scan = scan;

	record->superclass_specific_function_list
				= &mxs_struck_quick_scan_scan_function_list;

	record->class_specific_function_list = NULL;

	for ( i = 0; i < MXS_SQ_MAX_MOTORS; i++ ) {
		struck_quick_scan->motor_position_array[i] = NULL;
	}
	for ( i = 0; i < MXS_SQ_NUM_STRUCK_SCALERS; i++ ) {
		struck_quick_scan->data_array[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] =
		"mxs_struck_quick_scan_finish_record_initialization()";

	MX_SCAN *scan;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	long i;
	mx_status_type status;

	scan = (MX_SCAN *) record->record_superclass_struct;

	status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( ( scan->num_motors < 0 )
	  || ( scan->num_motors > MXS_SQ_MAX_MOTORS ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal num_motors value %ld for scan '%s'.  "
		"The allowed range is (0-%ld)",
			scan->num_motors, record->name,
			MXS_SQ_MAX_MOTORS );
	}

#if 0
	if ( scan->num_input_devices != 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error in scan '%s': "
		"There can only be one input device for a Struck quick scan, "
		"namely, the Struck MCS itself.",
			record->name );
	}
#endif

	for ( i = 0; i < scan->num_motors; i++ ) {
		scan->motor_is_independent_variable[i] = TRUE;
	}

	return mx_scan_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_dummy_function( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_prepare_for_scan_start()";

	MX_RECORD *struck_record;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long i, long_value;
	mx_status_type status;

	MX_DEBUG(-2,("%s invoked.", fname));

	status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Verify that the input device is a Struck MCS. */

	struck_record = ( scan->input_device_array )[0];

	if ( (struck_record->mx_superclass != MXR_DEVICE)
	  || (struck_record->mx_class != MXC_MULTICHANNEL_SCALER)
	  || (struck_record->mx_type != MXT_MCS_EPICS) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The input device '%s' is not a Struck MCS.  "
"Only a Struck MCS can be used as an input device for a Struck quick scan.",
			struck_record->name );
	}

	epics_mcs = (MX_EPICS_MCS *) struck_record->record_type_struct;

	if ( epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MCS pointer for record '%s' is NULL.",
			struck_record->name );
	}

	/* Get the EPICS record name for the Struck MCS. */

	mx_strncpy( struck_quick_scan->epics_record_name,
			epics_mcs->epics_record_name,
			sizeof( struck_quick_scan->epics_record_name ) );
 
	/* Allocate memory for the motor position array. */

	for ( i = 0; i < scan->num_motors; i++ ) {

		struck_quick_scan->motor_position_array[i] = (double *)
			malloc(quick_scan->num_measurements * sizeof(double));

		if ( struck_quick_scan->motor_position_array[i] == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an %ld array "
			"of motor positions for motor %ld.",
				quick_scan->num_measurements, i );
		}
	}

	/* Allocate memory for the scaler data array. */

	for ( i = 0; i < MXS_SQ_NUM_STRUCK_SCALERS; i++ ) {

		struck_quick_scan->data_array[i] = (long *)
			malloc(quick_scan->num_measurements * sizeof(long));

		if ( struck_quick_scan->data_array[i] == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an %ld array "
			"of data values for scaler %ld.",
				quick_scan->num_measurements, i );
		}
	}

	/* Initialize the datafile and plotting support. */

	status = mx_standard_prepare_for_scan_start( scan );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	if ( mx_plotting_is_enabled( scan->record ) ) {
		status = mx_plot_start_plot_section( &(scan->plot) );

		if ( status.code != MXE_SUCCESS ) {
			FREE_ARRAYS;
			return status;
		}
	}

	/* Send the motors to the start position for the scan. */

	MX_DEBUG(-2,("%s: Moving to start position.", fname));

	status = mx_motor_array_move_absolute(
			scan->num_motors,
			scan->motor_record_array,
			quick_scan->start_position,
			0 );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	/* Put the Struck EPICS record into MCS mode. */

	sprintf( pvname, "%s1.MODE", struck_quick_scan->epics_record_name );

	long_value = 1;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error putting Struck MCS '%s' (%s) into MCS mode.  EZCA status = %d (%s)",
			struck_record->name,
			struck_quick_scan->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		FREE_ARRAYS;
		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Set the channel advance source for the Struck MCS. */

	sprintf( pvname, "%s1.CHAS", struck_quick_scan->epics_record_name );

	long_value = 0;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the channel advance source for the Struck MCS '%s' (%s).  "
"EZCA status = %d (%s)",
			struck_record->name,
			struck_quick_scan->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		FREE_ARRAYS;
		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Set the prescale for the Struck MCS. */

	sprintf( pvname, "%s1.PSCL", struck_quick_scan->epics_record_name );

	long_value = 0;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the prescale for the Struck MCS '%s' (%s).  "
"EZCA status = %d (%s)",
			struck_record->name,
			struck_quick_scan->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		FREE_ARRAYS;
		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Wait for the motors to get to their start positions. */

	status = mx_wait_for_motor_array_stop( scan->num_motors,
				scan->motor_record_array, 0 );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	MX_DEBUG(-2,("%s: All motors are at the start position.", fname));

	/* Set the motor speeds for this quick scan. */

	for ( i = 0; i < scan->num_motors; i++ ) {

		status = mx_quick_scan_set_new_motor_speed( quick_scan, i );

		if ( status.code != MXE_SUCCESS ) {
			FREE_ARRAYS;
			return status;
		}
	}

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* Do not use FREE_ARRAYS in the execute_scan_body() function since
 * the cleanup_after_scan_end() function will attempt to use the arrays
 * regardless of what error code this routine returns.
 */

MX_EXPORT mx_status_type
mxs_struck_quick_scan_execute_scan_body( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_execute_scan_body()";

	MX_RECORD *struck_record;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status, direction;
	long i, j, start, read_cmd;
	long *data_ptr;
	unsigned long sleep_time;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_status;
	}

	struck_record = ( scan->input_device_array )[0];

	sleep_time = mx_round( 1000.0 * 
			mx_quick_scan_get_measurement_time(quick_scan) );

	/* Start the motors. */

	MX_DEBUG(-2,("%s: starting the motors.", fname));

	for ( i = 0; i < scan->num_motors; i++ ) {

		if ( quick_scan->end_position[i]
			>= quick_scan->start_position[i] )
		{
			direction = 1;
		} else {
			direction = -1;
		}

		MX_DEBUG(-2,("%s: starting '%s'", fname,
				scan->motor_record_array[i]->name ));

		mx_status = mx_motor_constant_velocity_move(
				scan->motor_record_array[i], direction );

		if ( mx_status.code != MXE_SUCCESS ) {
			for ( j = 0; j < i; j++ ) {
				(void) mx_motor_soft_abort(
					scan->motor_record_array[i] );
			}
			return mx_status;
		}
	}

	MX_DEBUG(-2,("%s: starting the MCS.", fname));

	/* Start the MCS. */

	sprintf( pvname, "%s1Start.VAL", struck_quick_scan->epics_record_name );

	start = 1;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &start );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Error starting the Struck MCS '%s' (%s).  EZCA status = %d (%s)",
			struck_record->name,
			struck_quick_scan->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Wait for the counting to finish. */

	sprintf( pvname, "%s1Start.VAL", struck_quick_scan->epics_record_name );

	do {
		ezca_status = ezcaGet( pvname, ezcaLong, 1, &start );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error setting the prescale for the Struck MCS '%s' (%s).  "
"EZCA status = %d (%s)",
				struck_record->name,
				struck_quick_scan->epics_record_name,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR, fname,
				mx_error_message);
		}
	} while ( start == 1 );

	MX_DEBUG(-2,("%s: Measurement complete.  Stopping the motors.",fname));

	/* Stop the motors. */

	for ( i = 0; i < scan->num_motors; i++ ) {
		(void) mx_motor_soft_abort( scan->motor_record_array[i] );
	}

	MX_DEBUG(-2,("%s: Reading out the MCS.", fname));

	/*  Get the data from the MCS channels. */

	for ( i = 0; i < MXS_SQ_NUM_STRUCK_SCALERS; i++ ) {

		sprintf( pvname, "%s%ld.READ",
			struck_quick_scan->epics_record_name, i );

		read_cmd = 1;

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &read_cmd );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error reading values for Struck MCS channel %ld.  EZCA status = %d (%s)",
				i, ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR, fname,
					mx_error_message);
		}

		sprintf( pvname, "%s%ld.VAL",
			struck_quick_scan->epics_record_name, i );

		data_ptr = struck_quick_scan->data_array[i];

		ezca_status = ezcaGet( pvname, ezcaLong, 
				quick_scan->num_measurements, data_ptr );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error transferring data for Struck MCS channel %ld.  EZCA status = %d (%s)",
				i, ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR, fname,
					mx_error_message);
		}
	}

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* Please note that most error conditions are ignored in the
 * cleanup_after_scan_end() function since it is important that
 * the data be saved to disk in spite of any extraneous errors
 * that may occur.
 */

MX_EXPORT mx_status_type
mxs_struck_quick_scan_cleanup_after_scan_end( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_cleanup_after_scan_end()";

	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	double motor_positions[ MXS_SQ_MAX_MOTORS ];
	long data_values[ MXS_SQ_NUM_STRUCK_SCALERS ];
	long i, j;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( mx_status.code != MXE_SUCCESS ) {

		/* Can't go any farther if this fails, so we abort. */

		FREE_ARRAYS;
		return mx_status;
	}

	MX_DEBUG(-2,("%s: restoring motor speeds.", fname));

	/* Restore the old motor speeds for this quick scan. */

	for ( i = 0; i < scan->num_motors; i++ ) {

		/* Errors in restoring the motor speed are ignored. */

		(void) mx_quick_scan_restore_old_motor_speed( quick_scan, i );
	}

	/* Copy the data in the scan data arrays to the datafile
	 * and send another copy to the plotting program.
	 */

	for ( i = 1; i < quick_scan->num_measurements; i++ ) {

		for ( j = 0; j < scan->num_motors; j++ ) {
			motor_positions[j] = 
				struck_quick_scan->motor_position_array[j][i];
		}
		for ( j = 0; j < scan->num_input_devices; j++ ) {
			data_values[j] = struck_quick_scan->data_array[j][i]
				- struck_quick_scan->data_array[j][i-1];
		}

		/* Skip duplicate points by looking at the timer value
		 * in input device channel 0.  If the timer value has not
		 * changed, skip the point.
		 */

		if ( data_values[0] == 0 ) {

			MX_DEBUG( 2,("%s: Discarding measurement %ld",
					fname, i));

			continue;  /* Go back to the top of the for() loop. */
		}

#if 0
		fprintf(stderr,"( %ld %ld ) ", scan->num_motors,
						scan->num_input_devices);

		for ( j = 0; j < scan->num_motors; j++ ) {
			fprintf(stderr," %g", motor_positions[j]);
		}
		for ( j = 0; j < MXS_SQ_NUM_STRUCK_SCALERS; j++ ) {
			fprintf(stderr," %ld", data_values[j]);
		}
		fprintf(stderr,"\n");
#endif

		MX_DEBUG( 2,("%s: Copying measurement %ld to data file.",
				fname, i));

		mx_status = mx_add_array_to_datafile( &(scan->datafile),
			MXFT_DOUBLE, scan->num_motors, motor_positions,
			MXFT_LONG, scan->num_input_devices, data_values );

		if ( mx_status.code != MXE_SUCCESS ) {

			/* If we cannot write the data to the datafile,
			 * all is lost, so we abort.
			 */

			FREE_ARRAYS;
			return mx_status;
		}

		if ( mx_plotting_is_enabled( scan->record ) ) {
			MX_DEBUG( 2,(
			    "%s: Adding measurement %ld to plot buffer.",
				fname, i));

			/* Failing to update the plot correctly is not
			 * a reason to abort since that would interrupt
			 * the writing of the datafile.
			 */

			(void) mx_add_array_to_plot_buffer( &(scan->plot),
			    MXFT_DOUBLE, scan->num_motors, motor_positions,
			    MXFT_LONG, scan->num_input_devices, data_values );
		}
	}

	if ( mx_plotting_is_enabled( scan->record ) ) {
		MX_DEBUG(-2,("%s: displaying the plot.", fname));

		/* Attempt to display the plot, but do not abort on an error.*/

		(void) mx_display_plot( &(scan->plot) );
	}

	/* Close the datafile and shut down the plot. */

	MX_DEBUG(-2,("%s: Invoking standard cleanup function.", fname));

	mx_status = mx_standard_cleanup_after_scan_end( scan );

	/* Free all of the data arrays that were allocated in the routine
	 * mxs_struck_quick_scan_prepare_for_scan_start().
	 */

	MX_DEBUG(-2,("%s: freeing the arrays.", fname));

	FREE_ARRAYS;

	MX_DEBUG(-2,("%s complete.", fname));

	return mx_status;
}

#endif /* HAVE_EPICS */

